#ifndef _APP_TYPES_H
#define _APP_TYPES_H

#include <stddef.h>

/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=4:tabstop=4:
 */
/*
 * Copyright (C) 2009 CEA/DAM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the CeCILL License.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license (http://www.cecill.info) and that you
 * accept its terms.
 */
/**
 * \file   lustre_hsm_types.h
 * \brief  Defines data types for Lustre-HSM purpose.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <errno.h>
#include "lustre_extended_types.h"
#include "global_config.h"

/********************************************************
 *       Types For Lustre-HSM Policy Engine             *
 ********************************************************/

#define FID_PK

typedef lustre_fid entry_id_t;

#define entry_id_equal( _p_id_1, _p_id_2 ) ( ((_p_id_1)->f_seq == (_p_id_2)->f_seq) && ((_p_id_1)->f_oid == (_p_id_2)->f_oid)  )


static int lmgr_gen_depth( void * tgt, const void * src )
{
    const char * fullpath = (char*)src;
    unsigned int * p_depth = (unsigned int *)tgt;

    const char     *curr;
    unsigned int   nb1;
    unsigned int   nb2;
    /* depth = number of '/' - 1 - depth of root fs.
     * E.g.: root="/mnt/lustre", path="/mnt/lustre/dir/foo", depth=4-2-1=1
     */

    nb1 = 0;
    curr = global_config.fs_path;
    while ( ( curr = strchr( curr, '/' ) ) )
    {
        curr++;
        nb1++;
    }

    nb2 = 0;
    curr = fullpath;
    while ( ( curr = strchr( curr, '/' ) ) )
    {
        curr++;
        nb2++;
    }

    (*p_depth) = nb2 - nb1 - 1;

   return 0;
}


typedef enum
{
  STATUS_UNKNOWN = 0,           /* undetermined status */
  STATUS_NEW,                   /* file has no HSM flags (just created) */
  STATUS_MODIFIED,              /* file must be archived */
  STATUS_RESTORE_RUNNING,       /* file is being retrieved */
  STATUS_ARCHIVE_RUNNING,       /* file is being archived */
  STATUS_SYNCHRO,               /* file has been synchronized in HSM, file can be purged */
  STATUS_RELEASED,              /* file is released (nothing to do). XXX should not be in DB? */
  STATUS_RELEASE_PENDING,    	/* file is being released */

  STATUS_COUNT                  /* number of possible file status */
} file_status_t;


typedef struct __entry_info__ 
{
	entry_id_t 	parent_id;
	char 	name[RBH_NAME_MAX];
	unsigned int 	path_update;
	char 	fullpath[RBH_PATH_MAX];
	unsigned int 	depth;
	unsigned int 	dircount;
	uint64_t 	avgsize;
	char 	owner[128];
	char 	gr_name[128];
	uint64_t 	size;
	uint64_t 	blocks;
	unsigned int 	creation_time;
	unsigned int 	last_access;
	unsigned int 	last_mod;
	char 	type[32];
	unsigned short 	mode;
	unsigned int 	nlink;
	file_status_t 	status;
	unsigned int 	md_update;
	int 	no_release;
	int 	no_archive;
	char 	archive_class[128];
	unsigned int 	arch_cl_update;
	char 	release_class[128];
	unsigned int 	rel_cl_update;
	stripe_info_t 	stripe_info;
	stripe_items_t 	stripe_items;
	unsigned int 	last_archive;
	unsigned int 	last_restore;
	char 	link[RBH_PATH_MAX];
} entry_info_t;

#define ATTR_INDEX_parent_id 	0
#define ATTR_INDEX_name 	1
#define ATTR_INDEX_path_update 	2
#define ATTR_INDEX_fullpath 	3
#define ATTR_INDEX_depth 	4
#define ATTR_INDEX_dircount 	5
#define ATTR_INDEX_avgsize 	6
#define ATTR_INDEX_owner 	7
#define ATTR_INDEX_gr_name 	8
#define ATTR_INDEX_size 	9
#define ATTR_INDEX_blocks 	10
#define ATTR_INDEX_creation_time 	11
#define ATTR_INDEX_last_access 	12
#define ATTR_INDEX_last_mod 	13
#define ATTR_INDEX_type 	14
#define ATTR_INDEX_mode 	15
#define ATTR_INDEX_nlink 	16
#define ATTR_INDEX_status 	17
#define ATTR_INDEX_md_update 	18
#define ATTR_INDEX_no_release 	19
#define ATTR_INDEX_no_archive 	20
#define ATTR_INDEX_archive_class 	21
#define ATTR_INDEX_arch_cl_update 	22
#define ATTR_INDEX_release_class 	23
#define ATTR_INDEX_rel_cl_update 	24
#define ATTR_INDEX_stripe_info 	25
#define ATTR_INDEX_stripe_items 	26
#define ATTR_INDEX_last_archive 	27
#define ATTR_INDEX_last_restore 	28
#define ATTR_INDEX_link 	29

#define ATTR_COUNT 30

#define ATTR_MASK_parent_id 	0X000001
#define ATTR_MASK_name 	0X000002
#define ATTR_MASK_path_update 	0X000004
#define ATTR_MASK_fullpath 	0X000008
#define ATTR_MASK_depth 	0X000010
#define ATTR_MASK_dircount 	0X000020
#define ATTR_MASK_avgsize 	0X000040
#define ATTR_MASK_owner 	0X000080
#define ATTR_MASK_gr_name 	0X000100
#define ATTR_MASK_size 	0X000200
#define ATTR_MASK_blocks 	0X000400
#define ATTR_MASK_creation_time 	0X000800
#define ATTR_MASK_last_access 	0X001000
#define ATTR_MASK_last_mod 	0X002000
#define ATTR_MASK_type 	0X004000
#define ATTR_MASK_mode 	0X008000
#define ATTR_MASK_nlink 	0X010000
#define ATTR_MASK_status 	0X020000
#define ATTR_MASK_md_update 	0X040000
#define ATTR_MASK_no_release 	0X080000
#define ATTR_MASK_no_archive 	0X100000
#define ATTR_MASK_archive_class 	0X200000
#define ATTR_MASK_arch_cl_update 	0X400000
#define ATTR_MASK_release_class 	0X800000
#define ATTR_MASK_rel_cl_update 	0X1000000
#define ATTR_MASK_stripe_info 	0X2000000
#define ATTR_MASK_stripe_items 	0X4000000
#define ATTR_MASK_last_archive 	0X8000000
#define ATTR_MASK_last_restore 	0X10000000
#define ATTR_MASK_link 	0X20000000

static const field_info_t field_infos[]=
{
	{ "parent_id", 	DB_ID, 	-1, 	FREQ_ACCESS | INDEXED | DNAMES, offsetof(entry_info_t, parent_id) , -1, NULL }, 
	{ "name", 	DB_TEXT, 	RBH_NAME_MAX-1, 	FREQ_ACCESS | DNAMES, offsetof(entry_info_t, name) , -1, NULL }, 
	{ "path_update", 	DB_UINT, 	-1, 	FREQ_ACCESS | DNAMES, offsetof(entry_info_t, path_update) , -1, NULL }, 
	{ "fullpath", 	DB_TEXT, 	RBH_PATH_MAX-1, 	FUNC_ATTR | DNAMES, offsetof(entry_info_t, fullpath) , -1, NULL }, 
	{ "depth", 	DB_UINT, 	-1, 	GENERATED, offsetof(entry_info_t, depth) , ATTR_INDEX_fullpath, lmgr_gen_depth }, 
	{ "dircount", 	DB_UINT, 	-1, 	DIR_ATTR, offsetof(entry_info_t, dircount) , ATTR_INDEX_type, NULL }, 
	{ "avgsize", 	DB_BIGUINT, 	-1, 	DIR_ATTR, offsetof(entry_info_t, avgsize) , ATTR_INDEX_type, NULL }, 
	{ "owner", 	DB_TEXT, 	127, 	FREQ_ACCESS, offsetof(entry_info_t, owner) , -1, NULL }, 
	{ "gr_name", 	DB_TEXT, 	127, 	FREQ_ACCESS, offsetof(entry_info_t, gr_name) , -1, NULL }, 
	{ "size", 	DB_BIGUINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, size) , -1, NULL }, 
	{ "blocks", 	DB_BIGUINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, blocks) , -1, NULL }, 
	{ "creation_time", 	DB_UINT, 	-1, 	INIT_ONLY, offsetof(entry_info_t, creation_time) , -1, NULL }, 
	{ "last_access", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, last_access) , -1, NULL }, 
	{ "last_mod", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, last_mod) , -1, NULL }, 
	{ "type", 	DB_ENUM_FTYPE, 	31, 	FREQ_ACCESS, offsetof(entry_info_t, type) , -1, NULL }, 
	{ "mode", 	DB_USHORT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, mode) , -1, NULL }, 
	{ "nlink", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, nlink) , -1, NULL }, 
	{ "status", 	DB_INT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, status) , -1, NULL }, 
	{ "md_update", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, md_update) , -1, NULL }, 
	{ "no_release", 	DB_BOOL, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, no_release) , -1, NULL }, 
	{ "no_archive", 	DB_BOOL, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, no_archive) , -1, NULL }, 
	{ "archive_class", 	DB_TEXT, 	127, 	FREQ_ACCESS, offsetof(entry_info_t, archive_class) , -1, NULL }, 
	{ "arch_cl_update", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, arch_cl_update) , -1, NULL }, 
	{ "release_class", 	DB_TEXT, 	127, 	FREQ_ACCESS, offsetof(entry_info_t, release_class) , -1, NULL }, 
	{ "rel_cl_update", 	DB_UINT, 	-1, 	FREQ_ACCESS, offsetof(entry_info_t, rel_cl_update) , -1, NULL }, 
	{ "stripe_info", 	DB_STRIPE_INFO, 	-1, 	INIT_ONLY, offsetof(entry_info_t, stripe_info) , -1, NULL }, 
	{ "stripe_items", 	DB_STRIPE_ITEMS, 	-1, 	INIT_ONLY, offsetof(entry_info_t, stripe_items) , -1, NULL }, 
	{ "last_archive", 	DB_UINT, 	-1, 	ANNEX_INFO, offsetof(entry_info_t, last_archive) , -1, NULL }, 
	{ "last_restore", 	DB_UINT, 	-1, 	ANNEX_INFO, offsetof(entry_info_t, last_restore) , -1, NULL }, 
	{ "link", 	DB_TEXT, 	RBH_PATH_MAX-1, 	ANNEX_INFO | SLINK_ATTR, offsetof(entry_info_t, link), -1, NULL }
};

#endif
